@@ -3,25 +3,27 @@
 import datetime
 from itertools import groupby
 
-from sql import operators, Literal, Null
-from sql.conditionals import Coalesce, Case
+from sql import Literal, Null, operators
+from sql.conditionals import Case, Coalesce
+from sql.functions import CharLength
 
 from trytond import backend
 from trytond.i18n import gettext
-from trytond.model import ModelSQL, ModelView, Workflow, fields, \
-        sequence_ordered
+from trytond.model import (
+    Index, ModelSQL, ModelView, Workflow, fields, sequence_ordered)
 from trytond.model.exceptions import AccessError
-from trytond.pool import Pool
-from trytond.pyson import Eval, If, Bool
-from trytond.transaction import Transaction
-from trytond.tools import sortable_values
-from trytond.wizard import Wizard, StateView, StateAction, StateTransition, \
-        Button
-
 from trytond.modules.company.model import (
-    employee_field, set_employee, reset_employee)
+    employee_field, reset_employee, set_employee)
+from trytond.modules.currency.fields import Monetary
 from trytond.modules.product import price_digits, round_price
-from .exceptions import InvoiceError
+from trytond.pool import Pool
+from trytond.pyson import Bool, Eval, If
+from trytond.tools import sortable_values
+from trytond.transaction import Transaction
+from trytond.wizard import (
+    Button, StateAction, StateTransition, StateView, Wizard)
+
+from .exceptions import InvalidRecurrence, InvoiceError
 
 
 class Subscription(Workflow, ModelSQL, ModelView):
@@ -30,29 +32,22 @@ class Subscription(Workflow, ModelSQL, ModelView):
     _rec_name = 'number'
 
     company = fields.Many2One(
-        'company.company', "Company", required=True, select=True,
+        'company.company', "Company", required=True,
         states={
-            'readonly': Eval('state') != 'draft',
+            'readonly': (Eval('state') != 'draft') | Eval('party', True),
             },
-        domain=[
-            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', -1)),
-            ],
-        depends=['state'],
         help="Make the subscription belong to the company.")
 
     number = fields.Char(
-        "Number", readonly=True, select=True,
+        "Number", readonly=True,
         help="The main identification of the subscription.")
-    # TODO revision
     reference = fields.Char(
-        "Reference", select=True,
+        "Reference",
         help="The identification of an external origin.")
     description = fields.Char("Description",
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
 
     party = fields.Many2One(
         'party.party', "Party", required=True,
@@ -60,22 +55,32 @@ class Subscription(Workflow, ModelSQL, ModelView):
             'readonly': ((Eval('state') != 'draft')
                 | (Eval('lines', [0]) & Eval('party'))),
             },
-        depends=['state'],
+        context={
+            'company': Eval('company', -1),
+            },
+        depends={'company'},
         help="The party who subscribes.")
     contact = fields.Many2One(
         'party.contact_mechanism', "Contact",
+        context={
+            'company': Eval('company', -1),
+            },
         search_context={
             'related_party': Eval('party'),
-            })
+            },
+        depends={'company', 'party'})
     invoice_party = fields.Many2One('party.party', "Invoice Party",
         states={
             'readonly': ((Eval('state') != 'draft')
                 | Eval('lines', [0])),
             },
+        context={
+            'company': Eval('company', -1),
+            },
         search_context={
             'related_party': Eval('party'),
             },
-        depends=['state', 'party'])
+        depends={'company', 'party'})
     invoice_address = fields.Many2One(
         'party.address', "Invoice Address",
         domain=[
@@ -85,30 +90,26 @@ class Subscription(Workflow, ModelSQL, ModelView):
         states={
             'readonly': Eval('state') != 'draft',
             'required': ~Eval('state').in_(['draft']),
-            },
-        depends=['party', 'invoice_party', 'state'])
+            })
     payment_term = fields.Many2One(
         'account.invoice.payment_term', "Payment Term",
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
 
     currency = fields.Many2One(
         'currency.currency', "Currency", required=True,
         states={
             'readonly': ((Eval('state') != 'draft')
                 | (Eval('lines', [0]) & Eval('currency', 0))),
-            },
-        depends=['state'])
+            })
 
     start_date = fields.Date(
         "Start Date", required=True,
         states={
             'readonly': ((Eval('state') != 'draft')
                 | Eval('next_invoice_date')),
-            },
-        depends=['state', 'next_invoice_date'])
+            })
     end_date = fields.Date(
         "End Date",
         domain=['OR',
@@ -120,31 +121,31 @@ class Subscription(Workflow, ModelSQL, ModelView):
             ],
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['start_date', 'state'])
+            })
 
     invoice_recurrence = fields.Many2One(
         'sale.subscription.recurrence.rule.set', "Invoice Recurrence",
         required=True,
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     invoice_start_date = fields.Date("Invoice Start Date",
         states={
             'readonly': ((Eval('state') != 'draft')
                 | Eval('next_invoice_date')),
-            },
-        depends=['state', 'next_invoice_date'])
-    next_invoice_date = fields.Date("Next Invoice Date", readonly=True)
+            })
+    next_invoice_date = fields.Date(
+        "Next Invoice Date", readonly=True,
+        states={
+            'invisible': Eval('state') != 'running',
+            })
 
     lines = fields.One2Many(
         'sale.subscription.line', 'subscription', "Lines",
         states={
             'readonly': ((Eval('state') != 'draft')
                 | ~Eval('start_date')),
-            },
-        depends=['state'])
+            })
 
     quoted_by = employee_field(
         "Quoted By", states=['quotation', 'running', 'closed', 'cancelled'])
@@ -156,12 +157,22 @@ class Subscription(Workflow, ModelSQL, ModelView):
             ('running', "Running"),
             ('closed', "Closed"),
             ('cancelled', "Cancelled"),
-            ], "State", readonly=True, required=True,
+            ], "State", readonly=True, required=True, sort=False,
         help="The current state of the subscription.")
 
     @classmethod
     def __setup__(cls):
+        cls.number.search_unaccented = False
+        cls.reference.search_unaccented = False
         super(Subscription, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(t, (t.reference, Index.Similarity())),
+                Index(
+                    t,
+                    (t.state, Index.Equality()),
+                    where=t.state.in_(['draft', 'quotation', 'running'])),
+                })
         cls._order = [
             ('start_date', 'DESC'),
             ('id', 'DESC'),
@@ -213,15 +224,22 @@ class Subscription(Workflow, ModelSQL, ModelView):
                 [table.state], ['cancelled'],
                 where=table.state == 'canceled'))
 
+    @classmethod
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [CharLength(table.number), table.number]
+
     @classmethod
     def default_company(cls):
         return Transaction().context.get('company')
 
     @classmethod
-    def default_currency(cls):
+    def default_currency(cls, **pattern):
         pool = Pool()
         Company = pool.get('company.company')
-        company = cls.default_company()
+        company = pattern.get('company')
+        if not company:
+            company = cls.default_company()
         if company:
             return Company(company).currency.id
 
@@ -229,14 +247,19 @@ class Subscription(Workflow, ModelSQL, ModelView):
     def default_state(cls):
         return 'draft'
 
-    @fields.depends('party', 'invoice_party')
+    @fields.depends('company', 'party', 'invoice_party', 'currency', 'lines')
     def on_change_party(self):
         if not self.invoice_party:
             self.invoice_address = None
+        if not self.lines:
+            self.currency = self.default_currency(
+                company=self.company.id if self.company else None)
         if self.party:
             if not self.invoice_party:
                 self.invoice_address = self.party.address_get(type='invoice')
             self.payment_term = self.party.customer_payment_term
+            if not self.lines and self.party.customer_currency:
+                self.currency = self.party.customer_currency
 
     @fields.depends('party', 'invoice_party')
     def on_change_invoice_party(self):
@@ -249,15 +272,15 @@ class Subscription(Workflow, ModelSQL, ModelView):
     @classmethod
     def set_number(cls, subscriptions):
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Config = pool.get('sale.configuration')
 
         config = Config(1)
         for subscription in subscriptions:
             if subscription.number:
                 continue
-            subscription.number = Sequence.get_id(
-                config.subscription_sequence.id)
+            subscription.number = config.get_multivalue(
+                'subscription_sequence',
+                company=subscription.company.id).get()
         cls.save(subscriptions)
 
     def compute_next_invoice_date(self):
@@ -267,7 +290,28 @@ class Subscription(Workflow, ModelSQL, ModelView):
         dt = datetime.datetime.combine(date, datetime.time())
         inc = (start_date == date) and not self.next_invoice_date
         next_date = rruleset.after(dt, inc=inc)
-        return next_date.date()
+        if next_date:
+            return next_date.date()
+
+    @classmethod
+    def validate_fields(cls, subscriptions, field_names):
+        super().validate_fields(subscriptions, field_names)
+        cls.validate_invoice_recurrence(subscriptions, field_names)
+
+    @classmethod
+    def validate_invoice_recurrence(cls, subscriptions, field_names=None):
+        if field_names and not (field_names & {
+                    'start_date', 'invoice_start_date', 'invoice_recurrence'}):
+            return
+        for subscription in subscriptions:
+            start_date = (
+                subscription.invoice_start_date or subscription.start_date)
+            try:
+                subscription.invoice_recurrence.rruleset(start_date)[0]
+            except IndexError:
+                raise InvalidRecurrence(gettext(
+                        'sale_subscription.msg_invoice_recurrence_invalid',
+                        subscription=subscription.rec_name))
 
     @classmethod
     def copy(cls, subscriptions, default=None):
@@ -283,7 +327,7 @@ class Subscription(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def view_attributes(cls):
-        return [
+        return super().view_attributes() + [
             ('/tree', 'visual',
                 If(Eval('state') == 'cancelled', 'muted', '')),
             ]
@@ -384,8 +428,6 @@ class Subscription(Workflow, ModelSQL, ModelView):
         all_invoice_lines = []
         for subscription, invoice in invoices.items():
             invoice_lines, _ = lines[subscription]
-            for line in invoice_lines:
-                line.invoice = invoice
             all_invoice_lines.extend(invoice_lines)
         InvoiceLine.save(all_invoice_lines)
 
@@ -424,6 +466,7 @@ class Subscription(Workflow, ModelSQL, ModelView):
             currency=self.currency,
             account=party.account_receivable_used,
             )
+        invoice.invoice_date = self.next_invoice_date
         invoice.on_change_type()
         invoice.payment_term = self.payment_term
         return invoice
@@ -434,13 +477,11 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
     __name__ = 'sale.subscription.line'
 
     subscription = fields.Many2One(
-        'sale.subscription', "Subscription", required=True, select=True,
-        ondelete='CASCADE',
+        'sale.subscription', "Subscription", required=True, ondelete='CASCADE',
         states={
             'readonly': ((Eval('subscription_state') != 'draft')
                 & Bool(Eval('subscription'))),
             },
-        depends=['subscription_state'],
         help="Add the line below the subscription.")
     subscription_state = fields.Function(
         fields.Selection('get_subscription_states', "Subscription State"),
@@ -463,21 +504,18 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         context={
             'company': Eval('company', None),
             },
-        depends=['subscription_state', 'company'])
+        depends={'company'})
     description = fields.Text("Description",
         states={
             'readonly': Eval('subscription_state') != 'draft',
-            },
-        depends=['subscription_state'])
+            })
 
     quantity = fields.Float(
-        "Quantity", digits=(16, Eval('unit_digits', 2)),
+        "Quantity", digits='unit',
         states={
             'readonly': Eval('subscription_state') != 'draft',
             'required': Bool(Eval('consumption_recurrence')),
-            },
-        depends=[
-            'unit_digits', 'subscription_state', 'consumption_recurrence'])
+            })
     unit = fields.Many2One(
         'product.uom', "Unit", required=True,
         states={
@@ -487,35 +525,29 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
             If(Bool(Eval('service_unit_category')),
                 ('category', '=', Eval('service_unit_category')),
                 ('category', '!=', -1)),
-            ],
-        depends=['subscription_state', 'service_unit_category'])
-    unit_digits = fields.Function(
-        fields.Integer("Unit Digits"), 'on_change_with_unit_digits')
+            ])
     service_unit_category = fields.Function(
         fields.Many2One('product.uom.category', "Service Unit Category"),
         'on_change_with_service_unit_category')
 
-    unit_price = fields.Numeric(
-        "Unit Price", digits=price_digits,
+    unit_price = Monetary(
+        "Unit Price", currency='currency', digits=price_digits,
         states={
             'readonly': Eval('subscription_state') != 'draft',
-            },
-        depends=['subscription_state'])
+            })
     currency = fields.Function(fields.Many2One('currency.currency',
         'Currency'), 'on_change_with_currency')
     consumption_recurrence = fields.Many2One(
         'sale.subscription.recurrence.rule.set', "Consumption Recurrence",
         states={
             'readonly': Eval('subscription_state') != 'draft',
-            },
-        depends=['subscription_state'])
+            })
     consumption_delay = fields.TimeDelta(
         "Consumption Delay",
         states={
             'readonly': Eval('subscription_state') != 'draft',
             'invisible': ~Eval('consumption_recurrence'),
-            },
-        depends=['subscription_state', 'consumption_recurrence'])
+            })
     next_consumption_date = fields.Date("Next Consumption Date", readonly=True)
     next_consumption_date_delayed = fields.Function(
         fields.Date("Next Consumption Delayed"),
@@ -529,9 +561,7 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         states={
             'readonly': ((Eval('subscription_state') != 'draft')
                 | Eval('consumed_until')),
-            },
-        depends=[
-            'subscription_start_date', 'subscription_state', 'consumed_until'])
+            })
     end_date = fields.Date(
         "End Date",
         domain=['OR', [
@@ -548,9 +578,12 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         states={
             'readonly': ((Eval('subscription_state') != 'draft')
                 | (~Eval('consumed_until') & Eval('consumed_until'))),
-            },
-        depends=['subscription_end_date', 'start_date',
-            'consumed_until', 'subscription_state', 'consumed_until'])
+            })
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('subscription')
 
     @classmethod
     def __register__(cls, module):
@@ -626,12 +659,6 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
     def default_quantity(cls):
         return 1
 
-    @fields.depends('unit')
-    def on_change_with_unit_digits(self, name=None):
-        if self.unit:
-            return self.unit.digits
-        return 2
-
     @fields.depends('subscription', '_parent_subscription.currency')
     def on_change_with_currency(self, name=None):
         if self.subscription and self.subscription.currency:
@@ -665,12 +692,11 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         category = product.sale_uom.category
         if not self.unit or self.unit.category != category:
             self.unit = product.sale_uom
-            self.unit_digits = product.sale_uom.digits
 
         with Transaction().set_context(self._get_context_sale_price()):
             self.unit_price = Product.get_sale_price(
                 [product], self.quantity or 0)[product.id]
-            if self.unit_price:
+            if self.unit_price is not None:
                 self.unit_price = round_price(self.unit_price)
 
         self.consumption_recurrence = self.service.consumption_recurrence
@@ -706,11 +732,9 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         pool = Pool()
         Lang = pool.get('ir.lang')
         lang = Lang.get()
-        return (lang.format(
-                '%.*f', (self.unit.digits, self.quantity or 0))
-            + '%s %s @ %s' % (
-                self.unit.symbol, self.service.rec_name,
-                self.subscription.rec_name))
+        return (lang.format_number_symbol(
+                self.quantity or 0, self.unit, digits=self.unit.digits)
+            + ' %s @ %s' % (self.service.rec_name, self.subscription.rec_name))
 
     @classmethod
     def search_rec_name(cls, name, clause):
@@ -793,12 +817,35 @@ class Line(sequence_ordered(), ModelSQL, ModelView):
         rruleset = self.consumption_recurrence.rruleset(self.start_date)
         dt = datetime.datetime.combine(date, datetime.time())
         inc = (self.start_date == date) and not self.next_consumption_date
-        next_date = rruleset.after(dt, inc=inc).date()
-        for end_date in [self.end_date, self.subscription.end_date]:
-            if end_date:
-                if next_date > end_date:
-                    return None
-        return next_date
+        next_date = rruleset.after(dt, inc=inc)
+        if next_date:
+            next_date = next_date.date()
+            for end_date in [self.end_date, self.subscription.end_date]:
+                if end_date:
+                    if next_date > end_date:
+                        return None
+            return next_date
+
+    @classmethod
+    def validate_fields(cls, lines, field_names):
+        super().validate_fields(lines, field_names)
+        cls.validate_consumption_recurrence(lines, field_names)
+
+    @classmethod
+    def validate_consumption_recurrence(cls, lines, field_names=None):
+        if field_names and not (field_names & {
+                    'consumption_recurrence', 'start_date'}):
+            return
+        for line in lines:
+            if line.consumption_recurrence:
+                try:
+                    line.consumption_recurrence.rruleset(line.start_date)[0]
+                except IndexError:
+                    raise InvalidRecurrence(gettext(
+                            'sale_subscription'
+                            '.msg_consumption_recurrence_invalid',
+                            line=line.rec_name,
+                            subscription=line.subscription.rec_name))
 
     @classmethod
     def copy(cls, lines, default=None):
@@ -817,13 +864,10 @@ class LineConsumption(ModelSQL, ModelView):
     __name__ = 'sale.subscription.line.consumption'
 
     line = fields.Many2One(
-        'sale.subscription.line', "Line", required=True, select=True,
-        ondelete='RESTRICT')
-    quantity = fields.Float(
-        "Quantity", digits=(16, Eval('unit_digits', 2)),
-        depends=['unit_digits'])
-    unit_digits = fields.Function(
-        fields.Integer("Unit Digits"), 'on_change_with_unit_digits')
+        'sale.subscription.line', "Line", required=True, ondelete='RESTRICT')
+    quantity = fields.Float("Quantity", digits='unit', required=True)
+    unit = fields.Function(fields.Many2One(
+            'product.uom', "Unit"), 'on_change_with_unit')
     date = fields.Date("Date", required=True)
     invoice_line = fields.Many2One(
         'account.invoice.line', "Invoice Line", readonly=True)
@@ -831,13 +875,13 @@ class LineConsumption(ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(LineConsumption, cls).__setup__()
+        cls.__access__.add('line')
         cls._order.insert(0, ('date', 'DESC'))
 
     @fields.depends('line')
-    def on_change_with_unit_digits(self, name=None):
+    def on_change_with_unit(self, name=None):
         if self.line and self.line.unit:
-            return self.line.unit.digits
-        return 2
+            return self.line.unit.id
 
     @classmethod
     def copy(cls, consumptions, default=None):
@@ -882,33 +926,18 @@ class LineConsumption(ModelSQL, ModelView):
                 consumptions, key=cls._group_invoice_key):
             sub_consumptions = list(sub_consumptions)
             line = InvoiceLine(**dict(key))
-            line.invoice_type = 'out'
+            line.invoice = invoice
+            line.on_change_invoice()
             line.type = 'line'
             line.quantity = sum(c.quantity for c in sub_consumptions)
+            line.on_change_product()
 
-            line.account = line.product.account_revenue_used
             if not line.account:
                 raise InvoiceError(
                     gettext('sale_subscription'
                         '.msg_consumption_invoice_missing_account_revenue',
                         product=line.product.rec_name))
 
-            taxes = []
-            pattern = line._get_tax_rule_pattern()
-            party = invoice.party
-            for tax in line.product.customer_taxes_used:
-                if party.customer_tax_rule:
-                    tax_ids = party.customer_tax_rule.apply(tax, pattern)
-                    if tax_ids:
-                        taxes.extend(tax_ids)
-                    continue
-                taxes.append(tax.id)
-            if party.customer_tax_rule:
-                tax_ids = party.customer_tax_rule.apply(None, pattern)
-                if tax_ids:
-                    taxes.extend(tax_ids)
-            line.taxes = taxes
-
             lines.append(line)
             grouped_consumptions.append(sub_consumptions)
         return lines, grouped_consumptions
